/* -*-c,save-*- */
/*
 * PATDEF.H - Pattern definations
 * Robert Heller. Created: Sat Oct 19, 1985 13:20:03.09
 * Last Mod: 
 * 
 * (c) Copyright 1985 by Robert Heller
 *     All Rights Reserved
 * 
 * 
 */
/*
 * This file contains the definations for the Pattern Matching System.
 * This includes typedefs, #defines and externs.
 */

/* a basic string descriptor: */
typedef struct {
    char *base;			/* base pointer to the string */
    short int offset;		/* offset to first character */
    short int length;		/* string length */
    } STRING_DESCR;

/* argument descriptor */
typedef struct arg_descr {
    short int data_type;	/* data type of argument */
#define AD_FLONUM 1		/* arg is a float */
#define AD_FIXNUM 2		/* arg is an int */
#define AD_STRING 3		/* arg is a pointer to a string descr */
#define AD_FUNCTION 4		/* arg is a pointer to a function */
#define AD_PATTERN 5		/* arg is a pointer to a pattern */
#define AD_UNDEFINED 0		/* undefined data type */
    union {
	float flonum;		/* float lives here */
	long int fixnum;	/* int lives here */
	STRING_DESCR *string;	/* string descr. ptr lives here */
	long int (*function)(); /* function ptr. lives here */
	struct pattern_node *pattern; /* and pattern lives here */
	void *pval;		/* Random Value pointer */
	} value;
    } ARG_DESCR;

/* pattern node object */
typedef struct pattern_node {
    int (*prog)();	/* match routine */
    struct pattern_node *subs;	/* pointer to sub. nodes */
    struct pattern_node *alts;	/* pointer to alt. nodes */
    ARG_DESCR *arg;		/* argument */
    short int resid;		/* residual */
    short int __mark;		/* mark field (used by copy functions) */
    struct pattern_node *__new;	/* pointer to new pattern (used by copy functions) */
    } PATTERN_NODE;

#ifndef PMODULE
#ifdef SUCCESS
#undef SUCCESS
#endif
#endif
extern PATTERN_NODE *breakk(STRING_DESCR *),*breakk_c(char *),*len(int);
extern PATTERN_NODE *any(STRING_DESCR *),*any_c(char *),*notany(STRING_DESCR *);
extern PATTERN_NODE *notany_c(char *),*span(STRING_DESCR *),*span_c(char *s);
extern PATTERN_NODE *arbno(PATTERN_NODE *);
extern PATTERN_NODE *alt(PATTERN_NODE *p1,PATTERN_NODE *p2);
extern PATTERN_NODE *concat(PATTERN_NODE *p1,PATTERN_NODE *p2);
extern PATTERN_NODE *lit_string(STRING_DESCR *),*c_lit_string(char *);
extern PATTERN_NODE *star(long int (*funct)());
extern PATTERN_NODE *cassign(PATTERN_NODE *,ARG_DESCR *);
extern PATTERN_NODE *assign(PATTERN_NODE *,ARG_DESCR *),*pos(int),*rpos(int);
extern PATTERN_NODE *tab(int),*rtab(int);
extern PATTERN_NODE *NIL,*FENCE,*FAIL,*SUCCESS,*ABORT,*REM,*ARB,*BAL;
/* define some macros to make it easier to allocate specific types */
#define icons(i) (acons(AD_FIXNUM,(void *)(long int) (i)))
#define fcons(f) (acons(AD_FLONUM,(void *)(float) (f)))
#define sacons(s)(acons(AD_STRING,(void *)(STRING_DESCR *) (s)))
#define pacons(p)(acons(AD_PATTERN,(void *)(PATTERN_NODE *)(p)))
#define fncons(fn)(acons(AD_FUNCTION,(void *)fn))
extern ARG_DESCR *acons(int type /*type code */,void * v /* value */);
extern void pattern_init();
extern int pmatch(char *subj,PATTERN_NODE *pat,STRING_DESCR *matched);
extern PATTERN_NODE *pncons(int (*p)(ARG_DESCR *),PATTERN_NODE *s,PATTERN_NODE *a,ARG_DESCR *ar,int r);
extern STRING_DESCR *scons(char *b,int o,int l);
extern STRING_DESCR *build_string(char *str);

#define MATCH_SUCCESS 1	/* success == TRUE */
#define MATCH_FAIL    0 /* failure == FALSE */

#ifdef OSK
#define perror prerr
#endif
