/* -*-c,save-*- */
/*
 * PATTERN.C - Pattern match functions
 * Robert Heller. Created: Sat Oct 19, 1985 13:52:15.52
 * Last Mod: Sun Oct 27, 1985 19:05:28.05
 * 
 * (c) Copyright 1985 by Robert Heller
 *     All Rights Reserved
 * 
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#ifdef STRING
#undef STRING
#endif

#define PMODULE
#include "patdef.h"
#define TRUE 1
#define FALSE 0
/*
 * global, "constant" primitives
 */

PATTERN_NODE *NIL,*FENCE,*FAIL,*SUCCESS,*ABORT,*REM,*ARB,*BAL;

static PATTERN_NODE _nil,_fence,_fail,_success,_abort,_rem,_arb1,_arb2,_arb3,
		    _gbal,_bal1,_bal2;


static void alt1(PATTERN_NODE *p1,PATTERN_NODE *p2);
static void clear_marks(PATTERN_NODE *son);
static void update_resid( PATTERN_NODE *son,int resupd);
static void concat1(PATTERN_NODE *son,PATTERN_NODE *nephew);
static PATTERN_NODE *copy_pat(PATTERN_NODE *son);
static int scan(int l,PATTERN_NODE *node);
static int m_break(ARG_DESCR *a);
static int m_span(ARG_DESCR *a);
static int m_any(ARG_DESCR *a);
static int m_notany(ARG_DESCR *a);
static int m_nil(ARG_DESCR *a);
static int m_success(ARG_DESCR *a);
static int m_fail(ARG_DESCR *a);
static int m_abort(ARG_DESCR *a);
static int m_fence(ARG_DESCR *a);
static int m_pos(ARG_DESCR *arg);
static int m_rpos(ARG_DESCR *arg);
static int m_rem(ARG_DESCR *a);
static int m_string(ARG_DESCR *arg);
static int m_len(ARG_DESCR *arg);
static int m_tab(ARG_DESCR *arg);
static int m_rtab(ARG_DESCR *arg);
static int m_arb1(ARG_DESCR *a);
static int m_arb2(ARG_DESCR *a);
static int m_gbal(ARG_DESCR *a);
static int m_star(ARG_DESCR *arg);
static int m_restar(ARG_DESCR *a);
static int m_va1(ARG_DESCR *a);
static int m_vab1(ARG_DESCR *a);
static int m_va2(ARG_DESCR *arg);
static int m_vab2(ARG_DESCR *a);
static int m_iva2(ARG_DESCR *a);
static void do_assign(long int snb,char *subjbase);
static int memchar(char c,STRING_DESCR *s);


/*
 *  pattern_init() - initialize global, "constant" primitives
 */
void pattern_init()
{
    PATTERN_NODE *temp;

    /* NIL primitive - match a zero length string */
    NIL = (&_nil);
    NIL->prog = m_nil;
    NIL->subs = NIL->alts = (PATTERN_NODE *) (NIL->arg = (ARG_DESCR *) 0L);
    NIL->resid = NIL->__mark = 0;

    /* FENCE - matchs the null string going forward, but aborts if backed
       into */
    FENCE = (&_fence);
    FENCE->prog = m_fence;
    FENCE->subs = FENCE->alts = (PATTERN_NODE *) (FENCE->arg = 
    												(ARG_DESCR *) 0L);
    FENCE->resid = FENCE->__mark = 0;

    /* FAIL - fails. */
    FAIL = (&_fail);
    FAIL->prog = m_fail;
    FAIL->subs = FAIL->alts = (PATTERN_NODE *) (FAIL->arg = (ARG_DESCR *) 0L);
    FAIL->resid = FAIL->__mark = 0;

    /* SUCCESS - succedes */
    SUCCESS = (&_success);
    SUCCESS->prog = m_success;
    SUCCESS->subs = SUCCESS->alts = (PATTERN_NODE *) (SUCCESS->arg = 
    													(ARG_DESCR *) 0L);
    SUCCESS->resid = SUCCESS->__mark = 0;

    /* ABORT - aborts */
    ABORT = (&_abort);
    ABORT->prog = m_abort;
    ABORT->subs = ABORT->alts = (PATTERN_NODE *) (ABORT->arg = 
    												(ARG_DESCR *) 0L);
    ABORT->resid = ABORT->__mark = 0;

    /* REM - matches the reaminder of the object string */
    REM = (&_rem);
    REM->prog = m_rem;
    REM->subs = REM->alts = (PATTERN_NODE *) (REM->arg = (ARG_DESCR *) 0L);
    REM->resid = REM->__mark = 0;

    /* ARB - matches an arbitrary string */
    /* (this one is a compound */
    ARB = (&_arb1);
    ARB->prog = m_arb1;
    ARB->alts = (PATTERN_NODE *) (ARB->arg = (ARG_DESCR *) 0L);
    ARB->resid = ARB->__mark = 0;
    ARB->subs = temp = (&_arb2);
    temp->prog = m_nil;
    temp->subs = (PATTERN_NODE *) (temp->arg = (ARG_DESCR *) 0L);
    temp->resid = temp->__mark = 0;
    temp->alts = (&_arb3);
    temp = temp->alts;
    temp->prog = m_arb2;
    temp->subs = temp->alts = (PATTERN_NODE *) (temp->arg = (ARG_DESCR *) 0L);
    temp->resid = temp->__mark = 0;

    /* BAL - matches a paren balanced string */
    BAL = (&_bal1);
    temp = (&_gbal);
    BAL->prog = m_nil;
    BAL->subs = temp;
    BAL->alts = (PATTERN_NODE *) (BAL->arg = (ARG_DESCR *) 0L);
    BAL->resid = 1;
    BAL->__mark = 0;
    temp->prog = m_gbal;
    temp->alts = (PATTERN_NODE *) (temp->arg = (ARG_DESCR *) 0L);
    temp->subs = (&_bal2);
    temp->resid = temp->__mark = 0;
    temp->subs->alts = temp;
    temp = temp->subs;
    temp->prog = m_nil;
    temp->subs = (PATTERN_NODE *) (temp->arg = (ARG_DESCR *) 0L);
    temp->resid = temp->__mark = 0;
    }
/*
 * STACK_SIZE is somewhat conservative
 */

#define STACK_SIZE 128

/* stacks */

static long int history_stack[STACK_SIZE];	/* history stack */
static struct nl_item {
    long int precur,postcur;
    ARG_DESCR *var;
    } namelist[STACK_SIZE];			/* name list stack */
static long int alpha_stack[STACK_SIZE];	/* alpha stack */

/* stack pointers */
static long int STACKPTR = -1L,
		NAMESP = -1L,
		ALPHASP = -1L,
		STACKBOT = -1L;

/* other "static" (free/bound vars) */

static STRING_DESCR *SUBJECT;	/* the current subject */
static long int LENGTH,		/* length */
		CURSOR,		/* current position */
		FUTILITY;	/* futility flag */
static PATTERN_NODE *NODE;	/* current node */


/*
 * stack functions
 */

/* push an item onto the history (main) stack */
static long int push(long int item)
{
    history_stack[++STACKPTR] = item;
    return(item);
    }
/* define a macro to make it easy */
#define HPUSH(X) (push((long int)(X)))
/* define a macro to the top of the stack. returns garbage if stack is 
   empty */
#define HTOP() (history_stack[STACKPTR])
/* pop an item off of the history stack */
static long int HPOP()
{
    if (STACKPTR <= STACKBOT) return(0L);
    else return(history_stack[STACKPTR--]);
    }
/* push an item onto alpha stack */
static long int apush(long int item)
{
    alpha_stack[++ALPHASP] = item;
    return(item);
    }
/* define a macro to make it easy */
#define APUSH(X) (apush((long int)(X)))
/* define a macro to the top of the stack. returns garbage if stack is 
   empty */
#define ATOP() (alpha_stack[ALPHASP])
/* pop an item off of the history stack */
static long int APOP()
{
    if (ALPHASP <= -1L) return(0L);
    else return(alpha_stack[ALPHASP--]);
    }
/* push an item onto the name list  stack */
static struct nl_item *NPUSH(long int prec,long int postc,ARG_DESCR *ap)
{
    NAMESP++;
    namelist[NAMESP].precur = prec;
    namelist[NAMESP].postcur = postc;
    namelist[NAMESP].var = ap;
    return(&namelist[NAMESP]);
    }
/* define a macro to the top of the stack. returns garbage if stack is 
   empty */
#define NTOP() (&namelist[NAMESP])
/* pop an item off of the name list stack */
static struct nl_item *NPOP()
{
    if (NAMESP <= -1L) return(0L);
    else return(&namelist[NAMESP--]);
    }
/*
 * memory funtions
 */
/* allocate a pattern node. */
PATTERN_NODE *pncons(int (*p)(ARG_DESCR *),/* prog field */
			    PATTERN_NODE *s,/* subs field */
			    PATTERN_NODE *a,/* alts field */
			    ARG_DESCR *ar,/* arg descr */
			    int r/* residual */)
{
    register PATTERN_NODE *new;

    /* allocate the space */
    new = (PATTERN_NODE *) calloc(sizeof(PATTERN_NODE),1);
    /* allocation failure? if so die */
    if (new == 0L) {
	perror("pncons");
	abort();
	}
    /* fill in fields */
    new->prog = p;
    new->subs = s;
    new->alts = a;
    new->arg = ar;
    new->resid = r;
    new->__mark = 0;
    return(new);
    }
/* allocate an arg descriptor */
ARG_DESCR *acons(int type /*type code */,void * v /* value */)
{
    register ARG_DESCR *new;

    /* allocate some memory */
    new = (ARG_DESCR *) calloc(sizeof(ARG_DESCR),1);
    /* if allocation failure, bomb out */
    if (new == 0L) {
	perror("acons");
	abort();
	}
    /* fill in fields */
    new->data_type = type;
    new->value.pval = v;
    return(new);
    }

/* allocate a string descriptor */
STRING_DESCR *scons(char *b,int o,int l)
{
    register STRING_DESCR *new;

    /* allocate some memory */
    new = (STRING_DESCR *) calloc(sizeof(STRING_DESCR),1);
    /* check for allocation failure */
    if (new == 0L) {
	perror("scons");
	abort();
	}
    /* fill is fields */
    new->base = b;
    new->offset = o;
    new->length = l;
    return(new);
    }
/* build a string descr from a string (string is copied) */
STRING_DESCR *build_string(char *str)
{
    register char *newstr;

    /* allocate some memory */
    newstr = calloc(strlen(str)+1,1);
    /* check for allocation failure */
    if (newstr == 0L) {
	perror("build_string");
	abort();
	}
    /* copy string */
    strcpy(newstr,str);
    /* return a string descr */
    return(scons(newstr,0,strlen(newstr)));
    }
/* alternation pattern builder function */
PATTERN_NODE *alt(PATTERN_NODE *p1,PATTERN_NODE *p2)
{

    p1 = copy_pat(p1);	/* copy pattern p1 */
    alt1(p1,p2);	/* alternate copy with p2 */
    return(p1);
    }
/* helper function */
static void alt1(PATTERN_NODE *p1,PATTERN_NODE *p2)
{
    register PATTERN_NODE *p;

    clear_marks(p1);
    for (p=p1; (p != 0L) && (p->alts != 0L) && (p->__mark <= 0); p = p->alts)
	p->__mark = 1;
    if (p != 0L && p->alts == 0L) p->alts = p2;
    }
/* pattern concatenation */
PATTERN_NODE *concat(PATTERN_NODE *p1,PATTERN_NODE *p2)
{
    p1 = copy_pat(p1);
    update_resid(p1,p2->resid);
    concat1(p1,p2);
    return(p1);
    }
#define FLD(p,i)(((i) == 1)?(p)->subs:(p)->alts)
#define SFLD(p,i,nv) if ((i) == 1) p->subs = nv; else p->alts = nv
/* helper routine */
static void concat1(PATTERN_NODE *son,PATTERN_NODE *nephew)
{
    register PATTERN_NODE *father,*gs,*gf;
    register int i;

    if (son == 0L) return;
    father = 0L;
    clear_marks(son);
    nephew->__mark = 1;
cc1_2:
    son->__mark = 1;
    if (son->subs == 0L) son->subs = nephew;
    i = 0;
cc1_1:
    i++;
    if (i>2) goto cc1_3;
    if (son == 0L) goto cc1_3;
    gs = FLD(son,i);
    if (gs == 0L) goto cc1_1;
    if (gs->__mark > 0) goto cc1_1;
    son->__mark = i;
    SFLD(son,i,father);
    father = son;
    son = gs;
    goto cc1_2;
cc1_3:
    if (father == 0L) return;
    i = father->__mark;
    gf = FLD(father,i);
    SFLD(father,i,son);
    son = father;
    father = gf;
    goto cc1_1;
    }
/* helper routine - smart deep copy */
static PATTERN_NODE *copy_pat(PATTERN_NODE *son)
{
    register PATTERN_NODE *father,*gs,*gf;
    register int i;

    if (son == 0L) return(son);
    father = 0L;
    clear_marks(son);
copy_2:
    son->__new = pncons(son->prog,son->subs,son->alts,son->arg,son->resid);
    son->__mark = 1;
    son = son->__new;
    i = 0;
copy_1:
    i++;
    if (i>2) goto copy_3;
    if (son == 0L) goto copy_3;
    gs = FLD(son,i);
    if (gs == 0L) goto copy_1;
    if (gs->__mark > 0) {
	SFLD(son,i,gs->__new);
	goto copy_1;
	}
    son->__mark = i;
    SFLD(son,i,father);
    father = son;
    son = gs;
    goto copy_2;
copy_3:
    if (father == 0L) return(son);
    i = father->__mark;
    gf = FLD(father,i);
    SFLD(father,i,son);
    son = father; father = gf;
    goto copy_1;
    }
/* helper function - clear marks */
static void clear_marks(PATTERN_NODE *son)
{
    register PATTERN_NODE *father,*gs,*gf;
    register int i;

    if (son == 0L) return;
    father = 0L;
clear_2:
    son->__mark = -1;
    i = 0;
clear_1:
    i++;
    if (i>2) goto clear_3;
    if (son == 0L) goto clear_3;
    gs = FLD(son,i);
    if (gs == 0L) goto clear_1;
    if (gs->__mark < 0) goto clear_1;
    son->__mark = -i;
    SFLD(son,i,father);
    father = son;
    son = gs;
    goto clear_2;
clear_3:
    if (father == 0L) return;
    i = -(father->__mark);
    gf = FLD(father,i);
    SFLD(father,i,son);
    son=father;
    father=gf;
    goto clear_1;
    }
/* helper routine - update resid */
static void update_resid( PATTERN_NODE *son,int resupd)
{
    register PATTERN_NODE *father,*gs,*gf;
    register int i;

    if (son == 0L) return;
    father = 0L;
    clear_marks(son);
upd_2:
    son->__mark = 1;
    son->resid += resupd;
    i = 0;
upd_1:
    i++;
    if (i>2) goto upd_3;
    if (son == 0L) goto upd_3;
    gs = FLD(son,i);
    if (gs == 0L) goto upd_1;
    if (gs->__mark > 0) goto upd_1;
    son->__mark = i;
    SFLD(son,i,father);
    father = son;
    son = gs;
    goto upd_2;
upd_3:
    if (father == 0L) return;
    i = father->__mark;
    gf = FLD(father,i);
    SFLD(father,i,son);
    son = father;
    father = gf;
    goto upd_1;
    }
/*
 * pattern constructor primitives
 */
/* breakk(str) - break primitive */
PATTERN_NODE *breakk(STRING_DESCR *str)
{
    register ARG_DESCR *argl;

    argl = sacons(str);
    return(pncons(m_break,0L,0L,argl,0));
    }
/* breakk_c(s) - simular, but s is simply a char ptr */
PATTERN_NODE *breakk_c(char *s)
{
    return(breakk(build_string(s)));
    }
/* span(str) - span primitive */
PATTERN_NODE *span(STRING_DESCR *str)
{
    register ARG_DESCR *argl;

    argl = sacons(str);
    return(pncons(m_span,0L,0L,argl,1));
    }
/* span_c(s) - simular, but s is simply a char ptr */
PATTERN_NODE *span_c(char *s)
{
    return(span(build_string(s)));
    }
/* any(str) - any primitive */
PATTERN_NODE *any(STRING_DESCR *str)
{
    register ARG_DESCR *argl;

    argl = sacons(str);
    return(pncons(m_any,0L,0L,argl,1));
    }
/* any_c(s) - simular, but s is simply a char ptr */
PATTERN_NODE *any_c(char *s)
{
    return(any(build_string(s)));
    }
/* notany(str) - notany primitive */
PATTERN_NODE *notany(STRING_DESCR *str)
{
    register ARG_DESCR *argl;

    argl = sacons(str);
    return(pncons(m_notany,0L,0L,argl,1));
    }
/* notany_c(s) - simular, but s is simply a char ptr */
PATTERN_NODE *notany_c(char *s)
{

    return(notany(build_string(s)));
    }
/* lit_string(str) - literal string primitive */
PATTERN_NODE *lit_string(STRING_DESCR *str)
{
    register ARG_DESCR *argl;

    argl = sacons(str);
    return(pncons(m_string,0L,0L,argl,str->length));
    }
/* c_lit_string(s) - simular, but s is simply a char ptr */
PATTERN_NODE *c_lit_string(char *s)
{

    return(lit_string(build_string(s)));
    }
/* len(l) - len primitive */
PATTERN_NODE *len(int l)
{
    register ARG_DESCR *argl;

    argl = icons(l);
    return(pncons(m_len,0L,0L,argl,l));
    }
/* pos(l) - pos primitive */
PATTERN_NODE *pos(int l)
{
    register ARG_DESCR *argl;

    argl = icons(l);
    return(pncons(m_pos,0L,0L,argl,0));
    }
/* rpos(l) - rpos primitive */
PATTERN_NODE *rpos(int l)
{
    register ARG_DESCR *argl;

    argl = icons(l);
    return(pncons(m_rpos,0L,0L,argl,0));
    }
/* tab(l) - tab primitive */
PATTERN_NODE *tab(int l)
{
    register ARG_DESCR *argl;

    argl = icons(l);
    return(pncons(m_tab,0L,0L,argl,0));
    }
/* rtab(l) - rtab primitive */
PATTERN_NODE *rtab(int l)
{
    register ARG_DESCR *argl;

    argl = icons(l);
    return(pncons(m_rtab,0L,0L,argl,0));
    }
/*
 * primitive variant compounds
 */
/* arbno(p) - compound to match an arbitary number (incl. 0) of what
 * pattern p matches */
PATTERN_NODE *arbno(PATTERN_NODE *p)
{
    register PATTERN_NODE *temp;

    temp = pncons(m_nil,0L,0L,0L,0);
    p = concat(p,temp);
    temp->alts = p;
    return(pncons(m_nil,temp,0L,0L,0));
    }
/* star(funct) - "unevaluated expr", implemented in C as a function call */
PATTERN_NODE *star(long int (*funct)())
{
    register ARG_DESCR *argl;
    register PATTERN_NODE *s1,*s2,*s3;

    argl = fncons(funct);
    s1 = pncons(m_star,0L,0L,argl,0);
    s2 = pncons(m_nil,0L,0L,0L,0);
    s3 = pncons(m_restar,0L,0L,0L,0);
    s1->subs = s2;
    s3->subs = s2;
    s2->alts = s3;
    return(s1);
    }
/* cassign(p,v) - conditionall assign string match by pattern p to "var" v */
PATTERN_NODE *cassign(PATTERN_NODE *p,ARG_DESCR *v)
{
    PATTERN_NODE *pncons(),*concat();
    register PATTERN_NODE *v1,*v2,*vb1,*vb2;

    vb2 = pncons(m_vab2,0L,0L,0L,0);
    v2 = pncons(m_va2,0L,vb2,v,0);
    v2 = concat(p,v2);
    vb1 = pncons(m_vab1,0L,0L,0L,0);
    v1 = pncons(m_va1,v2,vb1,0L,v2->resid);
    return(v1);
    }
/* assign(p,v) - unconditionall assign string match by pattern p to "var" v */
PATTERN_NODE *assign(PATTERN_NODE *p,ARG_DESCR *v)
{
    register PATTERN_NODE *v1,*v2,*vb1;

    v2 = pncons(m_iva2,0L,0L,v,0);
    v2 = concat(p,v2);
    vb1 = pncons(m_vab1,0L,0L,0L,0);
    v1 = pncons(m_va1,v2,vb1,0L,v2->resid);
    return(v1);
    }
/*
 * pattern match functions
 */

/* pattern match status codes */
#define SCAN_SUCCESS	0	/* success */
#define SCAN_L_FAIL	1	/* length failure */
#define SCAN_M_FAIL	2	/* match failure */
#define SCAN_ABORT	3	/* forced abort */

/* user entry point: match subj against pat. fill match with info about 
 * sub-string matched */
int pmatch(subj,pat,matched)
register char *subj;
register PATTERN_NODE *pat;
register STRING_DESCR *matched;
{
    register int status;
    register long int precursor;
    long int sspb,ssp,sab,snb,scur,sfut;
    STRING_DESCR subject;
    STRING_DESCR *ssubj;

    sspb = STACKBOT;
    ssp = STACKPTR;
    sab = ALPHASP;
    snb = NAMESP;
    sfut = FUTILITY;
    scur = CURSOR;
    ssubj = SUBJECT;
    STACKBOT = STACKPTR;
    SUBJECT = &subject;
    subject.base = subj;
    subject.offset = 0;
    subject.length = strlen(subj);

    status = SCAN_L_FAIL;
    for (precursor = 0; precursor<subject.length; precursor++) {
	CURSOR = precursor;
	STACKPTR = STACKBOT;
	ALPHASP = sab;
	NAMESP = snb;
	FUTILITY = TRUE;
	HPUSH(0);
	HPUSH(precursor);
	status = scan(subject.length,pat);
	if (status == SCAN_SUCCESS || status == SCAN_ABORT || FUTILITY) break;
	}
    if (status == SCAN_SUCCESS) {
	matched->base = subj;
	matched->offset = precursor;
	matched->length = CURSOR-precursor;
	do_assign(snb,subj);
	}
    STACKBOT = sspb;
    STACKPTR = ssp;
    ALPHASP = sab;
    NAMESP = snb;
    FUTILITY = sfut;
    CURSOR = scur;
    SUBJECT = ssubj;
    return((status == SCAN_SUCCESS)?MATCH_SUCCESS:MATCH_FAIL);
    }
/* helper function - scan: this is where all the work is done */
static int scan(int l,PATTERN_NODE *node)
{
    PATTERN_NODE *snode;
    long int slen;
    register int status;
    long int HPOP();
/*    int pattern_init();
    register long int pbase,fun;*/

/*    pbase = (long int) pattern_init;*/
    snode = NODE;
    NODE = node;
    slen = LENGTH;
    LENGTH = l;
    status = SCAN_SUCCESS;

    while(NODE != 0L) {
/*	fun = (long int) NODE->prog;
	fun -= pbase;
	printf("*** In scan(): NODE = %08lx\n->prog = %08lx\n->subs = %08lx\n",
	       NODE,fun,NODE->subs);
	printf("->alts = %08lx\n->arg = %08lx\n->resid = %d\n->__mark = %d\n",
	       NODE->alts,NODE->arg,NODE->resid,NODE->__mark);
	printf("->__new = %08ld\n",NODE->__new);
	printf("CURSOR = %ld, LENGTH = %ld, FUTILITY = %ld\nSTACKPTR = %ld\n",
	       CURSOR,LENGTH,FUTILITY,STACKPTR);*/
/*	if (NODE->arg != 0L) {
	    switch (NODE->arg->data_type) {
	    case FIXNUM: 
		printf("arg is FIXNUM: %ld\n",NODE->arg->value.fixnum);
		break;
	    case FLONUM:
		printf("arg is FLONUM: %10.5f\n",NODE->arg->value.flonum);
		break;
	    case STRING:
		printf("arg is STRING: '");
		{ int i,l; char *c;
		    l = NODE->arg->value.string->length;
		    c = NODE->arg->value.string->base+
			NODE->arg->value.string->offset;
		    for (i=0;i<l;i++) printf("%c",*c++);
		    }
		printf("'\n");
		break;
	    case FUNCTION:
		printf("arg is FUNCTION: %08lx\n",NODE->arg->value.function);
		break;
	    case PATTERN:
		printf("arg is PATTERN: %08lx\n",NODE->arg->value.pattern);
		break;
	    case UNDEFINED:
		printf("arg is UNDEFINED\n");
		break;
		}
	    }*/
	if (NODE->alts != 0L) {
	    HPUSH(NODE->alts);
	    HPUSH(CURSOR);
	    }
	status = (*(NODE->prog))(NODE->arg);
/*	printf("status is %d\n",status);*/
	switch (status) {
	case SCAN_SUCCESS:
	    NODE = NODE->subs;
	    break;
	case SCAN_M_FAIL: FUTILITY = FALSE;
	case SCAN_L_FAIL:
	    CURSOR = HPOP();
	    NODE = (PATTERN_NODE *) HPOP();
	    break;
	case SCAN_ABORT: goto scan_exit;
	    }
	}
scan_exit:
    NODE = snode;
    LENGTH = slen;
    return(status);
    }
/*
 * primitives
 */
/* break primitive */
static int m_break(ARG_DESCR *a)
{
    register STRING_DESCR *s;

    s = a->value.string;
    for (;CURSOR<LENGTH;CURSOR++)
	if (memchar(*(SUBJECT->base+SUBJECT->offset+CURSOR),s))
	    return(SCAN_SUCCESS);
    return(SCAN_L_FAIL);
    }
/* span primitive */
static int m_span(a)
register ARG_DESCR *a;
{
    register STRING_DESCR *s;
    register long int sc;

    s = a->value.string;
    sc = CURSOR;
    if (CURSOR == LENGTH) return(SCAN_L_FAIL);
    for (;CURSOR<LENGTH;CURSOR++)
	if (!memchar(*(SUBJECT->base+SUBJECT->offset+CURSOR),s)) break;
    if (CURSOR == sc) return(SCAN_M_FAIL);
    else return(SCAN_SUCCESS);
    }
/* any primitive */
static int m_any(a)
register ARG_DESCR *a;
{
    register STRING_DESCR *s;

    s = a->value.string;
    if (CURSOR == LENGTH) return(SCAN_L_FAIL);
    else if (memchar(*(SUBJECT->base+SUBJECT->offset+CURSOR),s)) {
	CURSOR++;
	return(SCAN_SUCCESS);
	}
    else return(SCAN_M_FAIL);
    }
/* notany primitive */
static int m_notany(a)
register ARG_DESCR *a;
{
    register STRING_DESCR *s;

    s = a->value.string;
    if (CURSOR == LENGTH) return(SCAN_L_FAIL);
    else if (!memchar(*(SUBJECT->base+SUBJECT->offset+CURSOR),s)) {
	CURSOR++;
	return(SCAN_SUCCESS);
	}
    else return(SCAN_M_FAIL);
    }
/* nil primitive */
static int m_nil(ARG_DESCR * a)
{
    return(SCAN_SUCCESS);
    }
/* success primitive */
static int m_success(ARG_DESCR * a)
{
    return(SCAN_SUCCESS);
    }
/* fail primitive */
static int m_fail(ARG_DESCR *a)
{
    return(SCAN_M_FAIL);
    }
/* abort primitive */
static int m_abort(ARG_DESCR *a)
{
    return(SCAN_ABORT);
    }
/* fence primitive */
static int m_fence(ARG_DESCR *a)
{
    static PATTERN_NODE abt;
    static int init = FALSE;

    if (!init) {
	abt.prog = m_abort;
	abt.alts = abt.subs = (PATTERN_NODE *) (abt.arg = (ARG_DESCR *) 0L);
	abt.resid = 0;
	}
    HPUSH(&abt);
    HPUSH(CURSOR);
    return(SCAN_SUCCESS);
    }
/* pos primitive */
static int m_pos(ARG_DESCR *arg)
{
    register long int p;

    p = arg->value.fixnum;
    if (p < 0 || p>LENGTH) {
	FUTILITY = TRUE;
	return(SCAN_L_FAIL);
	}
    else if (p==CURSOR) return(SCAN_SUCCESS);
    else if (p<CURSOR) return(SCAN_M_FAIL);
    else {
	FUTILITY = TRUE;
	return(SCAN_L_FAIL);
	}
    }
/* rpos primitive */
static int m_rpos(ARG_DESCR *arg)
{
    ARG_DESCR a1;

    a1.data_type = AD_FIXNUM;
    a1.value.fixnum = LENGTH-arg->value.fixnum;
    if (a1.value.fixnum < 0) {
	FUTILITY = TRUE;
	return(SCAN_L_FAIL);
	}
    return(m_pos(&a1));
    }
/* rem primitive */
static int m_rem(ARG_DESCR *a)
{
    CURSOR = LENGTH;
    return(SCAN_SUCCESS);
    }
/* literal string primitive */
static int m_string(ARG_DESCR *arg)
{
    register STRING_DESCR *s;
    register long int l;
    register char *sc;

    s = arg->value.string;
    if ((s->length+CURSOR)>LENGTH) return(SCAN_L_FAIL);
    for (l=s->length,sc=s->base+s->offset;
	 l > 0 && *sc == *(SUBJECT->base+SUBJECT->offset+CURSOR);
	 sc++,l--,CURSOR++) ;
    if (l != 0) return(SCAN_M_FAIL);
    else return(SCAN_SUCCESS);
    }
/* len primitive */
static int m_len(ARG_DESCR *arg)
{
    register long int l;

    l = arg->value.fixnum;
    if (CURSOR+l > LENGTH) return(SCAN_L_FAIL);
    else {
	CURSOR += l;
	return(SCAN_SUCCESS);
	}
    }
/* tab primitive */
static int m_tab(ARG_DESCR *arg)
{
    register long int p;

    p = arg->value.fixnum;
    if (p < 0 || p>LENGTH) {
	FUTILITY = TRUE;
	return(SCAN_L_FAIL);
	}
    else if (p<=CURSOR) {
	CURSOR = p;
	return(SCAN_SUCCESS);
	}
    else {
	FUTILITY = TRUE;
	return(SCAN_L_FAIL);
	}
    }
/* rtab primitive */
static int m_rtab(ARG_DESCR *arg)
{
    ARG_DESCR a1;

    a1.data_type = AD_FIXNUM;
    a1.value.fixnum = LENGTH-arg->value.fixnum;
    if (a1.value.fixnum < 0) {
	FUTILITY = TRUE;
	return(SCAN_L_FAIL);
	}
    return(m_tab(&a1));
    }
/* arb1 primitive (part of ARB) */
static int m_arb1(ARG_DESCR *a)
{
   HPUSH(FUTILITY);
   FUTILITY = TRUE;
   return(SCAN_SUCCESS);
   }
/* arb2 primitive (part of ARB) */
static int m_arb2(ARG_DESCR *a)
{
    if (FUTILITY) {
	FUTILITY = HPOP();
	return(SCAN_L_FAIL);
	}
    else if ((CURSOR++) <= LENGTH) {
	HPOP();
	return(SCAN_M_FAIL);
	}
    else {
	HPUSH(NODE);
	HPUSH(CURSOR);
	return(SCAN_SUCCESS);
	}
    }
/* gbal primitive (part of BAL) */
static int m_gbal(ARG_DESCR *a)
{
    register int paren_count;

    if (CURSOR == LENGTH) return(SCAN_L_FAIL);
    if (*(SUBJECT->base+SUBJECT->offset+CURSOR) == '(') paren_count = 1;
    else if  (*(SUBJECT->base+SUBJECT->offset+CURSOR) == ')') 
	return(SCAN_M_FAIL);
    else paren_count = 0;
    CURSOR++;
    while (CURSOR <= LENGTH && paren_count != 0) {
	if (*(SUBJECT->base+SUBJECT->offset+CURSOR) == '(') paren_count++;
	else if  (*(SUBJECT->base+SUBJECT->offset+CURSOR) == ')') 
	    paren_count--;
	CURSOR++;
	}
    if (paren_count == 0) return(SCAN_SUCCESS);
    else return(SCAN_M_FAIL);
    }
/* star primitive (part of star()) */
static int m_star(ARG_DESCR *arg)
{
    register ARG_DESCR *p;
    PATTERN_NODE *lit_string();
    static char tempst[20];
    register int reduction,stat,allflg;
    register PATTERN_NODE *pp;
    
    p = arg;
    while (p != 0L && p->data_type == AD_FUNCTION) 
	p = (ARG_DESCR *) (*(p->value.function))(p);
    if (p == 0L) return(SCAN_M_FAIL);
    if (p->data_type == AD_PATTERN) {
	allflg = 0;
	pp = p->value.pattern;
	}
    else if (p->data_type == AD_STRING) {
	allflg = 1;
	pp = lit_string(p->value.string);
	}
    else {
	allflg = 2;
	if (p->data_type == AD_FIXNUM) sprintf(tempst,"%ld",p->value.fixnum);
	else sprintf(tempst,"%f20.10",p->value.flonum);
	pp = c_lit_string(tempst);
	}
    HPUSH(0L); HPUSH(CURSOR);
    reduction = NODE->resid;
    if ((reduction+CURSOR) > LENGTH) {
	if (allflg == 1) free(pp);
	else if (allflg == 2) {
	    free(pp->arg->value.string->base);
	    free(pp->arg->value.string);
	    free(pp->arg);
	    free(pp);
	    }
	return(SCAN_L_FAIL);
	}
    stat = scan((int) (LENGTH - reduction),pp);
    if (allflg == 1) free(pp);
    else if (allflg == 2) {
	free(pp->arg->value.string);
	free(pp);
	}
    if (stat == SCAN_SUCCESS) return(SCAN_SUCCESS);
    else return(SCAN_M_FAIL);
    }
/* restar primitive (part of star()) */
static int m_restar(ARG_DESCR *a)
{
    register PATTERN_NODE *pp;
    register int reduction,stat;

    CURSOR = HPOP();
    pp = (PATTERN_NODE *) HPOP();
    if (pp == 0L) {
	return(SCAN_M_FAIL);
	}
    else {
	reduction = NODE->resid;
	if ((reduction+CURSOR) > LENGTH) return(SCAN_L_FAIL);
	stat = scan((int) (LENGTH - reduction),pp);
	if (stat == SCAN_SUCCESS) return(SCAN_SUCCESS);
	else return(SCAN_M_FAIL);
	}
    }
/* va1 primitive - part of cassign and assign */
static int m_va1(ARG_DESCR *a)
{
    APUSH(CURSOR);
/*    printf("*** In m_va1(): CURSOR (%ld) pushed, ALPHASP=%ld\n",CURSOR,ALPHASP);*/
    return(SCAN_SUCCESS);
    }
/* vab1 primitive - part of cassign and assign */
static int m_vab1(ARG_DESCR *a)
{
    APOP();
/*    printf("*** In m_vab1(): Alpha Stack popped, ALPHASP=%ld\n",ALPHASP);*/
    return(SCAN_M_FAIL);
    }
/* va2 primitive - part of cassign */
static int m_va2(arg)
register ARG_DESCR *arg;
{
    register long int precurs;
    long int APOP();

    precurs = APOP();
    NPUSH(precurs,CURSOR,arg);
/*    printf("*** In m_va2(): Alpha stack popped, precurs=%ld & CURSOR=%ld\n",
	   precurs,CURSOR);
    printf("                Name stack pushed.  ALPHASP=%ld, NAMESP=%ld\n",
	   ALPHASP,NAMESP);*/
    return(SCAN_SUCCESS);
    }
/* vab2 - part of cassign */
static int m_vab2(ARG_DESCR *a)
{
    register struct nl_item *nl;
    struct nl_item *NPOP();

    nl = NPOP();
    APUSH(nl->precur);
/*    printf("*** In m_vab2(): Name stack popped (precur = %ld). NAMESP=%ld,ALPHASP = %ld\n",
    	nl->precur,NAMESP,ALPHASP);*/
    return(SCAN_M_FAIL);
    }
/* iva2 primitive - part of assign */
static int m_iva2(ARG_DESCR *a)
{
    register long int prec;
    long int APOP();

    prec = APOP();
/*    printf("*** In m_iva2(): Alpha stack popped, prec=%ld, ALPHASP=%ld\n",
	prec,ALPHASP);*/
    while (a != 0L && a->data_type == AD_FUNCTION) 
	a = (ARG_DESCR *) (*(a->value.function))(a);
/*    printf("                 (after eval) a = %08lx,a->data_type=%d\n",a,
	(a==0L)?UNDEFINED:a->data_type);*/
    if (a == 0L) {
	APUSH(prec);
/*	printf("                 (a == 0L). Alpha stack pushed.\n");*/
	return(SCAN_M_FAIL);
	}
    if (a->data_type == AD_STRING) {
/*	printf("                 a->value.string = %08lx\n",a->value.string);*/
	a->value.string->base = SUBJECT->base;
	a->value.string->offset = SUBJECT->offset+prec;
	a->value.string->length = CURSOR - prec;
	}
    return(SCAN_SUCCESS);
    }
/*****************/
/*
 * general helper functions
 */
/* memchar - return TRUE if c is in s */
static int memchar(char c,STRING_DESCR *s)
{
    register char *ss;
    register int i,l;

    ss = s->base + s->offset;
    l = s->length;
    for (i=0;i<l;i++)
	if (c == *ss++) return(TRUE);
    return(FALSE);
    }
/* do_assign - process assignments */
static void do_assign(long int snb,char *subjbase)
{
    register struct nl_item *nl;
    register ARG_DESCR *v;
    register STRING_DESCR *s;
    struct nl_item *NPOP();
    static STRING_DESCR nullstr = {"",0,0};
    static ARG_DESCR anull; /* = {STRING,&nullstr}; */
    static int anull_init = FALSE;

	if (!anull_init) {
		anull.data_type = AD_STRING;
		anull.value.string = &nullstr;
		anull_init = TRUE;
	}
    while (NAMESP > snb) {
	nl = NPOP();
	v = nl->var;
	while (v != 0L &&
	       v->data_type == AD_FUNCTION) 
	    v = (ARG_DESCR *) (*(v->value.function))(&anull);
	if (v != 0L && v->data_type == AD_STRING) {
	    s = v->value.string;
	    s->base = subjbase;
	    s->offset = nl->precur;
	    s->length = nl->postcur - nl->precur;
	    }
	}
    }
/********************/
/* debug functions */
ARG_DESCR *DBG_p_stat()
{
    register char *p;
    register long int i;
    static ARG_DESCR null /* = {PATTERN,&_success} */;
    static int init = FALSE;
    
    if (!init) {
	null.data_type = AD_PATTERN;
	null.value.pattern = &_success;
	init = TRUE;
	}
    printf("*** In DBG_p_stat: CURSOR = %ld, LENGTH = %ld, FUTILITY = %ld\n",
	   CURSOR,LENGTH,FUTILITY);
    printf("*** SUBJECT = |");
    p = SUBJECT->base+SUBJECT->offset;
    for (i=0;i<CURSOR;i++) {
	printf("%c",*p);
	p++;
	}
    printf("\n               ");
    for (i=0;i<CURSOR;i++) printf("%c",' ');
    for (i=CURSOR;i<LENGTH;i++) {
	printf("%c",*p);
	p++;
	}
    printf("|\n");
    return(&null);
    }
